#include "tclInt.h"
#include "tclCompile.h"
#include "tclAssembly.h"
#include "tclOOInt.h"


static basicBlock * AllocBB(unsigned char *, int,int); 
static int AdvanceIp(const unsigned char *pc);
static int StackCheckBasicBlock(StackCheckerState* , basicBlock *, basicBlock *, int);
static int CheckStack(basicBlock *);

/*
 * Macro that encapsulates an efficiency trick that avoids a function call for
 * the simplest of compiles. The ANSI C "prototype" for this macro is:
 *
 * static void		CompileWord(CompileEnv *envPtr, Tcl_Token *tokenPtr,
 *			    Tcl_Interp *interp, int word);
 */

#define CompileWord(envPtr, tokenPtr, interp, word) \
    if ((tokenPtr)->type == TCL_TOKEN_SIMPLE_WORD) {			\
	TclEmitPush(TclRegisterNewLiteral((envPtr), (tokenPtr)[1].start, \
		(tokenPtr)[1].size), (envPtr));				\
    } else {								\
	envPtr->line = mapPtr->loc[eclIndex].line[word];		\
	envPtr->clNext = mapPtr->loc[eclIndex].next[word];		\
	TclCompileTokens((interp), (tokenPtr)+1, (tokenPtr)->numComponents, \
		(envPtr));						\
    }

/*
 * TIP #280: Remember the per-word line information of the current command. An
 * index is used instead of a pointer as recursive compilation may reallocate,
 * i.e. move, the array. This is also the reason to save the nuloc now, it may
 * change during the course of the function.
 *
 * Macro to encapsulate the variable definition and setup.
 */

#define DefineLineInformation \
    ExtCmdLoc *mapPtr = envPtr->extCmdMapPtr;				\
    int eclIndex = mapPtr->nuloc - 1

#define SetLineInformation(word) \
    envPtr->line = mapPtr->loc[eclIndex].line[(word)];			\
    envPtr->clNext = mapPtr->loc[eclIndex].next[(word)]

#define PushVarNameWord(i,v,e,f,l,s,sc,word) \
    PushVarName(i,v,e,f,l,s,sc,						\
	    mapPtr->loc[eclIndex].line[(word)],				\
	    mapPtr->loc[eclIndex].next[(word)])

/*
 * Flags bits used by PushVarName.
 */

#define TCL_NO_LARGE_INDEX 1	/* Do not return localIndex value > 255 */

#define HASNSQUALIFIERS(endlabel)   \
do { \
for (i = 0, p = instArgs[0];  i < instArgsLengths[0];  i++, p++) { \
         if ((*p == ':') && (i < instArgsLengths[0]-1) && (*(p+1) == ':')) { \
             result = TCL_ERROR; /* Name is not local, has :: in it */\
             goto endlabel;  \
         } \
     } \
 } \
while(0)   

#define EMITRIGHTOPCODE(endlabel,a,b,c,d) \
do { \
    if (localVar < 0) { \
        result = TCL_ERROR; /* The varible refrence is not a local */ \
        goto endlabel;  \
    } else if (localVar <= 255) { \
        TclEmitInstInt1((isLoad ? (a) : (b)), localVar, envPtr); \
    } else { \
        TclEmitInstInt4((isLoad ? (c) : (d)), localVar, envPtr); \
    }  \
} \
while(0)

#define EMITRIGHTJUMP(jumpKind) \
do { \
   switch(talInstructionTable[tblind].instType) { \
       case ASSEM_JUMP: \
           if ((jumpKind)) { \
               if (jumpDist > 127) { \
            	    TclEmitInstInt4(INST_JUMP4, jumpDist, envPtr); \
               } else { \
            	    TclEmitInstInt1(INST_JUMP1, jumpDist, envPtr); \
               } \
           } \
           else { \
               TclEmitInstInt4(INST_JUMP4, jumpDist, envPtr); \
           } \
           break; \
       case ASSEM_JUMP_FALSE: \
           if ((jumpKind)) { \
               if (jumpDist > 127) { \
           	       TclEmitInstInt4(INST_JUMP_FALSE4, jumpDist, envPtr); \
               } else { \
           	       TclEmitInstInt1(INST_JUMP_FALSE1, jumpDist, envPtr); \
               } \
           } \
           else { \
               TclEmitInstInt4(INST_JUMP_FALSE4, jumpDist, envPtr); \
           } \
           break; \
       case ASSEM_JUMP_TRUE: \
           if ((jumpKind)) { \
               if (jumpDist > 127) { \
                   TclEmitInstInt4(INST_JUMP_TRUE4, jumpDist, envPtr); \
               } else { \
                   TclEmitInstInt1(INST_JUMP_TRUE1, jumpDist, envPtr); \
               } \
           } \
           else { \
                    TclEmitInstInt4(INST_JUMP_TRUE4, jumpDist, envPtr); \
           } \
           break; \
       default: \
           break; \
   } \
} \
while (0)

#define UpdateBlockMinMaxStackDepth()   \
    if(envPtr->currStackDepth > curr_bb->maxStackDepth) { \
        curr_bb->maxStackDepth = envPtr->currStackDepth;  \
    }                                                     \
    if(envPtr->currStackDepth < curr_bb->minStackDepth) { \
        curr_bb->minStackDepth = envPtr->currStackDepth;  \        
    }

#define UpdateBlockFinalStackDepth()    \
    curr_bb->finalStackDepth = envPtr->currStackDepth; 
    
#define EndCurrBB_NoJump()   \
    curr_bb->may_fall_thru = 1;  \
    curr_bb->nextBlock = envPtr->codeNext; \
    curr_bb->jumpTarget = NULL 
 
#define EndCurrBB_Jump() \
    curr_bb->may_fall_thru = 0;  \
    curr_bb->nextBlock = envPtr->codeNext; 

talInstDesc talInstructionTable[] = {
    {"push",    +1  ,   ASSEM_PUSH  ,   INST_PUSH1  ,   0   ,   1},
    {"pop",     -1  ,   ASSEM_1BYTE ,   INST_POP    ,   1   ,   0},
    {"dup",     +1  ,   ASSEM_1BYTE ,   INST_DUP    ,   0   ,   1}, 
    {"lor",     -1  ,   ASSEM_1BYTE ,   INST_LOR    ,   2   ,   1},
    {"land",    -1  ,   ASSEM_1BYTE ,   INST_LAND   ,   2   ,   1},
    {"bitor",   -1  ,   ASSEM_1BYTE ,   INST_BITOR  ,   2   ,   1},
    {"bitxor",  -1  ,   ASSEM_1BYTE ,   INST_BITXOR ,   2   ,   1},
    {"bitand",  -1  ,   ASSEM_1BYTE ,   INST_BITAND ,   2   ,   1},
    {"eq",      -1  ,   ASSEM_1BYTE ,   INST_EQ     ,   2   ,   1},
    {"neq",     -1  ,   ASSEM_1BYTE ,   INST_NEQ    ,   2   ,   1},
    {"lt",      -1  ,   ASSEM_1BYTE ,   INST_LT     ,   2   ,   1},
    {"gt",      -1  ,   ASSEM_1BYTE ,   INST_GT     ,   2   ,   1},
    {"le",      -1  ,   ASSEM_1BYTE ,   INST_LE     ,   2   ,   1},
    {"ge",      -1  ,   ASSEM_1BYTE ,   INST_GE     ,   2   ,   1},
    {"lshift",  -1  ,   ASSEM_1BYTE ,   INST_LSHIFT ,   2   ,   1},
    {"rshift",  -1  ,   ASSEM_1BYTE ,   INST_RSHIFT ,   2   ,   1},
    {"add",     -1  ,   ASSEM_1BYTE ,   INST_ADD    ,   2   ,   1},
    {"sub",     -1  ,   ASSEM_1BYTE ,   INST_SUB    ,   2   ,   1},
    {"mult",    -1  ,   ASSEM_1BYTE ,   INST_MULT   ,   2   ,   1},
    {"div", -1, ASSEM_1BYTE, INST_DIV,2,1},
    {"mod", -1, ASSEM_1BYTE, INST_MOD,2,1},
    {"uplus", 0, ASSEM_1BYTE, INST_UPLUS,2,1},
    {"uminus", 0, ASSEM_1BYTE, INST_UMINUS,2,1},
    {"bitnot", 0, ASSEM_1BYTE, INST_BITNOT,2,1},
    {"not", 0, ASSEM_1BYTE, INST_LNOT,2,1},
    {"expon", -1, ASSEM_1BYTE, INST_EXPON,2,1},
    {"load", 1, ASSEM_LOAD, INST_LOAD_SCALAR1, 0,1}, 
    {"loadArray", 0, ASSEM_LOAD_ARR, INST_LOAD_ARRAY1, 1,1},
    {"loadStk", 0, ASSEM_LOAD_STK, INST_LOAD_SCALAR_STK,1,1},
    {"loadArrayStk", -1, ASSEM_LOAD_ARR_STK, INST_LOAD_ARRAY_STK, 2,1},
    {"store", 0, ASSEM_STORE, INST_STORE_SCALAR1, 1,1}, 
    {"storeArray", 0, ASSEM_STORE_ARR, INST_STORE_ARRAY1, 2,1}, 
    {"storeStk", -1, ASSEM_STORE_STK, INST_STORE_SCALAR_STK, 2,1}, 
    {"storeArrayStk", -2, ASSEM_STORE_ARR_STK, INST_STORE_ARRAY_STK, 3,1}, 
    {"incr", 0, ASSEM_INCR, INST_INCR_SCALAR1, 1,1},
    {"incrImm", 1, ASSEM_INCR_IMM, INST_INCR_SCALAR1_IMM, 2,1},
    {"incrStk", -1, ASSEM_INCR_STK, INST_INCR_SCALAR_STK, 2,1},
    {"incrStkImm" , 0, ASSEM_INCR_STK_IMM, INST_INCR_SCALAR_STK_IMM, 2,1},
    {"incrArray", -1, ASSEM_INCR_ARRAY, INST_INCR_ARRAY1, 3,1},
    {"incrArrayImm", 0, ASSEM_INCR_ARRAY_IMM, INST_INCR_ARRAY1_IMM, 3,1},
    {"incrArrayStk", -2, ASSEM_INCR_ARRAY_STK, INST_INCR_ARRAY_STK, 3,1},
    {"incrArrayStkImm",-1, ASSEM_INCR_ARRAY_STK_IMM, INST_INCR_ARRAY_STK_IMM, 3,1},    
    {"streq", -1, ASSEM_STR_EQ, INST_STR_EQ, 2,1},
    {"strneq", -1, ASSEM_STR_NEQ, INST_STR_NEQ, 2,1},
    {"strcmp", -1, ASSEM_STR_CMP, INST_STR_CMP, 2,1},
    {"strlen", 0, ASSEM_STR_LEN, INST_STR_LEN, 1,1},
    {"strindex", -1, ASSEM_STR_INDEX, INST_STR_INDEX, 2,1},
    {"strmatch", -1, ASSEM_STR_MATCH, INST_STR_MATCH, 2,1},
    {"listIndex", -1, ASSEM_LIST_INDEX, INST_LIST_INDEX ,2,1},
    {"listLength", 0, ASSEM_LIST_LENGTH, INST_LIST_LENGTH, 1,1},
    {"lsetList", 1, ASSEM_LSETLIST, INST_LSET_LIST, 3,1},
    {"over", 1, ASSEM_OVER, INST_OVER, 0,1},
    {"label", 0, ASSEM_LABEL, 0, 0,0}, 
    {"jump", 0, ASSEM_JUMP, INST_JUMP1,0,0},
    {"jumpTrue", -1, ASSEM_JUMP_TRUE, INST_JUMP_TRUE1, 1,0},
    {"jumpFalse", -1, ASSEM_JUMP_FALSE, INST_JUMP_FALSE1, 1,0},
    {NULL, 0, 0,0}
};

int
Tcl_AssembleObjCmd(
    ClientData dummy,		/* Not used. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *const objv[])	/* Argument objects. */
{
    if (objc != 2) {
    	Tcl_WrongNumArgs(interp, 1, objv, "bytecodeList");
    	return TCL_ERROR;
    }
    
    if (strlen(TclGetString(objv[1])) > 0) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj(
		    "bytecodeList should be \"\"", -1));
	    Tcl_SetErrorCode(interp, "TCL", "ARGUMENT", "FORMAT", NULL);
	    return TCL_ERROR;
    }

    return TCL_OK;
    
}

int TclCompileAssembleCmd(
    Tcl_Interp *interp,		/* Used for error reporting. */
    Tcl_Parse *parsePtr,	/* Points to a parse structure for the command
				 * created by Tcl_ParseCommand. */
    Command *cmdPtr,		/* Points to defintion of command being
				 * compiled. */
    CompileEnv *envPtr)		/* Holds resulting instructions. */
{ 
    Tcl_Obj *bcList;
    Tcl_Token *tokenPtr;
    tokenPtr = TokenAfter(parsePtr->tokenPtr);

    
    if (parsePtr->numWords != 2) {
        // Tcl_WrongNumArgs(interp, 1, NULL, "bytecodeList");
        // printf("right here because the numWords is %d\n",parsePtr->numWords);
	    return TCL_ERROR;
    }
    if (TclCheckByteCodeSyntax(interp, tokenPtr) == TCL_ERROR) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("bytecodeList should be a list of <bytecode, bytecodearg>", -1));
        Tcl_SetErrorCode(interp, "TCL", "ARGUMENT", "FORMAT", NULL);
        return TCL_ERROR;
    }
    
    bcList = Tcl_NewStringObj(tokenPtr[1].start, tokenPtr[1].size);
    
    if ((tokenPtr->type == TCL_TOKEN_SIMPLE_WORD)) {
        return TclAssembleCode(interp, bcList, envPtr, 0);
    }
    else if (tokenPtr->type != TCL_TOKEN_SIMPLE_WORD) {
        return TCL_ERROR;
    }
}

int TclAssembleCode(Tcl_Interp *interp,		/* Used for error reporting. */
  Tcl_Obj * bcList,
  CompileEnv *envPtr, 	/* Holds resulting instructions. */
  int direct)	
{
     int ind, instNameLength=0, instArgsLengths[2],bcListSize = 0, bcSize = 0;
     Tcl_Obj *bc, *inst, *arg;
     Tcl_Token *tokenPtr;

     char * instName; 
     char * instArgs[2];
     int i, localVar, localVarName = 1, isLoad = 0, result = TCL_OK,
        incrArray = 0, opnd = 0;
     label *l;
     Tcl_HashEntry * entry;
     const char *p;
     unsigned char * prev_inst;
     
     DefineLineInformation;	/* TIP #280 */
     
     if (Tcl_ListObjLength(interp, bcList, &bcListSize) == TCL_ERROR) {
         return TCL_ERROR;
     }
  
     if (bcListSize == 0) {
         int litIndex = TclRegisterNewLiteral(envPtr, "puts", -1);
         TclEmitPush(litIndex, envPtr);
         litIndex = TclRegisterNewLiteral(envPtr, "Empty Bytecode", -1);
         TclEmitPush(litIndex, envPtr);
         return TCL_OK;
     }
     else {
         Tcl_HashTable LabelHash, LabelIndHash, *LabelHashPtr, *LabelIndHashPtr, BBHash, *BBHashPtr;
         LabelHashPtr = &LabelHash;
         LabelIndHashPtr = &LabelIndHash;
         BBHashPtr = &BBHash;
          
         Tcl_InitHashTable(LabelHashPtr, TCL_STRING_KEYS);
         Tcl_InitHashTable(LabelIndHashPtr, TCL_ONE_WORD_KEYS);
         Tcl_InitHashTable(BBHashPtr, TCL_STRING_KEYS);
         basicBlock * curr_bb = 0, * head_bb = 0;
         if (curr_bb == 0) {
             curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
             head_bb = curr_bb;
         }
                  
         for (ind = 0; ind < bcListSize; ind++) {
             result = TCL_OK;
             Tcl_ListObjIndex(interp, bcList, ind, &bc);
             Tcl_ListObjLength(interp, bc, &bcSize);
             
             Tcl_ListObjIndex(interp, bc, 0, &inst);
             instName = Tcl_GetStringFromObj(inst, NULL);
             
             for (i=1; i < bcSize; i++) {
                 Tcl_ListObjIndex(interp, bc, i, &arg);
                 instArgs[i-1] = Tcl_GetStringFromObj(arg,&instArgsLengths[i-1]);
             }

             int tblind = 0;
             int litIndex;
             if (Tcl_GetIndexFromObjStruct(interp, inst, &talInstructionTable[0].name,
                 sizeof(talInstDesc), "msg", TCL_EXACT, &tblind) == TCL_ERROR) {
                    printf("Instruction \"%s\" could not be found.", instName);
                    return TCL_ERROR;
             }
            
             switch(talInstructionTable[tblind].instType) {
                case ASSEM_LABEL:
                    /* Add the (label_name, address) pair to the hash table */
                    entry = Tcl_CreateHashEntry(LabelHashPtr, instArgs[0], &opnd);
                    if (opnd != 0) {
                        l = (label *)ckalloc(sizeof(label));
                        l->isDefined = 1;
                        l->offset = envPtr->codeNext; 
                        Tcl_SetHashValue(entry, l);
                    }
                    else {  /* Label may have already been defined */
                        l = (label *) Tcl_GetHashValue(entry);
                        if (l->isDefined) {
                            printf("Label has already been defined\n");
                            result = TCL_ERROR;
                            goto end_of_else;
                        }
                        else {
                            /* Walk the previous jumps and correct them */      
                            unsigned char * temp;
                            unsigned char *prevJump =  TclGetInt4AtPtr(l->offset + 1);
                            TclStoreInt4AtPtr(envPtr->codeNext-l->offset, l->offset + 1);
                            while (prevJump != -1) {
                                temp = TclGetInt4AtPtr(prevJump + 1);
                                TclStoreInt4AtPtr(envPtr->codeNext, prevJump + 1);
                                prevJump = temp;
                            }
                            l->offset = envPtr->codeNext;
                            l->isDefined = 1;
                        }   
                    }
                    
                    if (curr_bb != 0) {
                        /* End this block and make a new one */ 
                        /* Allocate the new basic block */
                        curr_bb->may_fall_thru = 1;
                        curr_bb->successor1 = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                        UpdateBlockFinalStackDepth()
                        curr_bb = curr_bb->successor1;
                    }
                    else {
                        /* This is the first instruction 
                            start a new basic block */
                        curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                        head_bb = curr_bb;
                    }
                    /* Add the (label_name, basic_block_struct_address) pair to the hash table  */
                    entry = Tcl_CreateHashEntry(BBHashPtr, instArgs[0], &opnd);
                    Tcl_SetHashValue(entry, curr_bb);
                    
                    break;
                
                case ASSEM_JUMP_FALSE:
                case ASSEM_JUMP_TRUE:
                case ASSEM_JUMP: 
                    /*  Assumes the label is a string for now */
                    entry = Tcl_FindHashEntry(LabelHashPtr, instArgs[0]);
                    int jumpDist;
                    if (entry != NULL) {
                        l = (label *) Tcl_GetHashValue(entry);
                        if (l->isDefined) {
                            jumpDist = l->offset - envPtr->codeNext;
                            EMITRIGHTJUMP(1);
                        }
                        else {
                            jumpDist = l->offset;
                            l->offset = envPtr->codeNext;
                            EMITRIGHTJUMP(0);
                        }   
                    } /* Label not in hash */
                    else {
                        l = (label *)ckalloc(sizeof(label));
                        l->isDefined = 0;
                        l->offset = envPtr->codeNext;
                        entry = Tcl_CreateHashEntry(LabelHashPtr, instArgs[0], &opnd);
                        Tcl_SetHashValue(entry, l);
                        jumpDist = -1;
                        EMITRIGHTJUMP(0);
                    }
                    
                    /* Check if there is a basic block struct, JUMP may very well
                    be the first instruction. */
                    if (curr_bb != 0) {
                        if(talInstructionTable[tblind].instType == ASSEM_JUMP) {
                            curr_bb->jumpTargetLabelHashEntry = Tcl_CreateHashEntry(BBHashPtr, instArgs[0], &opnd);
                            curr_bb->may_fall_thru = 0;
                            curr_bb->successor1 = NULL;
                            UpdateBlockFinalStackDepth()
                            curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);

                        }
                        else {
                            curr_bb->jumpTargetLabelHashEntry = Tcl_CreateHashEntry(BBHashPtr, instArgs[0], &opnd);
                            curr_bb->may_fall_thru = 1;
                            curr_bb->successor1 = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                            UpdateBlockFinalStackDepth()
                            curr_bb = curr_bb->successor1;
                        }                        
                    }
                    break;
                    
                 
                case ASSEM_PUSH:
                    /* This may be the first intruction so we may have to create the 
                        initial basic block structure, also see LOAD */
                    if (curr_bb == 0) {
                        curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                        head_bb = curr_bb;
                    }
                    /* Must be simple word */
                    litIndex = TclRegisterNewLiteral(envPtr, instArgs[0], instArgsLengths[0]);
                    TclEmitPush(litIndex, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_1BYTE:
                    TclEmitOpcode(talInstructionTable[tblind].tclInstCode, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_LOAD:
                    isLoad = 1;
                    /* This may be the first intruction so we may have to create the 
                        initial basic block structure, also see PUSH */
                    if (curr_bb == 0) {
                        curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                        head_bb = curr_bb;
                    }
                case ASSEM_STORE:
                    /* Check to see if the name has namespace qualifiers, goto end if it does */
                    HASNSQUALIFIERS(load_store_end);
                     /* Maybe need to check if the name has array subscripting */   
                    localVar = TclFindCompiledLocal(instArgs[0], instArgsLengths[0], localVarName, envPtr);
                    EMITRIGHTOPCODE(load_store_end, INST_LOAD_SCALAR1, INST_STORE_SCALAR1, INST_LOAD_SCALAR4, INST_STORE_SCALAR4);
                    load_store_end: 
                    isLoad = 0;
                    if (result == TCL_ERROR) goto end_of_else;
                    UpdateBlockMinMaxStackDepth();
                    break;
                
                case ASSEM_LOAD_ARR:
                    isLoad = 1;
                case ASSEM_STORE_ARR:
                    /* Check to see if the name has namespace qualifiers, goto end if it does */
                    HASNSQUALIFIERS(load_arr_end); 
                    localVar = TclFindCompiledLocal(instArgs[0], instArgsLengths[0], localVarName, envPtr);
                    EMITRIGHTOPCODE(load_arr_end, INST_LOAD_ARRAY1, INST_STORE_ARRAY1, INST_LOAD_ARRAY4, INST_STORE_ARRAY4);
                    load_arr_end:
                    isLoad = 0;
                    if (result == TCL_ERROR) goto end_of_else;
                    UpdateBlockMinMaxStackDepth();
                    break;
                    
                case ASSEM_LOAD_STK:
                    isLoad = 1;
                case ASSEM_STORE_STK:
                    TclEmitOpcode((isLoad ? INST_LOAD_SCALAR_STK : INST_STORE_SCALAR_STK), envPtr);
                    isLoad = 0;
                    UpdateBlockMinMaxStackDepth();
                    break;
                    
                case ASSEM_LOAD_ARR_STK:
                    isLoad = 1;
                case ASSEM_STORE_ARR_STK:
                    TclEmitOpcode((isLoad ? INST_LOAD_ARRAY_STK : INST_STORE_ARRAY_STK), envPtr);
                    isLoad = 0;
                    UpdateBlockMinMaxStackDepth();
                    break;
                 
                case ASSEM_INCR_ARRAY:
                    incrArray = 1;
                case ASSEM_INCR:
                    HASNSQUALIFIERS(incr_end); 
                    localVar = TclFindCompiledLocal(instArgs[0], instArgsLengths[0], localVarName, envPtr);
                    if (localVar < 0 || localVar > 255) {
                        result = TCL_ERROR;
                        goto incr_end;
                    }
                    TclEmitInstInt1((incrArray ? INST_INCR_ARRAY1 : INST_INCR_SCALAR1), localVar, envPtr);
                    incr_end:
                    incrArray = 0;
                    if (result == TCL_ERROR) goto end_of_else;
                    UpdateBlockMinMaxStackDepth();
                    break;
                
                case ASSEM_INCR_ARRAY_IMM:
                    incrArray = 1;
                case ASSEM_INCR_IMM:
                    if (curr_bb == 0) {
                        curr_bb = AllocBB(envPtr->codeNext, 0, envPtr->currStackDepth);
                        head_bb = curr_bb;
                    }
                    HASNSQUALIFIERS(incr_imm_end); 
                    localVar = TclFindCompiledLocal(instArgs[0], instArgsLengths[0], localVarName, envPtr);
                    opnd = atoi(instArgs[1]);
                    if (opnd == 0 || opnd < -127 || opnd > 127 || localVar < 0 || localVar > 255) {
                        result = TCL_ERROR;
                        goto incr_imm_end;
                    }
                    TclEmitInstInt1((incrArray ? INST_INCR_ARRAY1_IMM : INST_INCR_SCALAR1_IMM), localVar, envPtr);
        		    TclEmitInt1(opnd, envPtr);
                    incr_imm_end:
                    incrArray = 1;
                    if (result == TCL_ERROR) goto end_of_else;
                    UpdateBlockMinMaxStackDepth();
                    break;
                
                case ASSEM_INCR_ARRAY_STK:
                    incrArray = 1;
                case ASSEM_INCR_STK:
                    TclEmitOpcode((incrArray ? INST_INCR_ARRAY_STK : INST_INCR_SCALAR_STK), envPtr);
                    incrArray = 0;
                    UpdateBlockMinMaxStackDepth();
                    break;
                
                case ASSEM_INCR_ARRAY_STK_IMM:
                    incrArray = 1;
                case ASSEM_INCR_STK_IMM:
                    opnd = atoi(instArgs[0]);
                    if (opnd == 0 || opnd < -127 || opnd > 127) {
                        result = TCL_ERROR;
                        goto incr_stk_imm_end;
                    }
                    TclEmitInstInt1((incrArray ? INST_INCR_ARRAY_STK_IMM : INST_INCR_SCALAR_STK_IMM), opnd, envPtr);
                    incr_stk_imm_end:
                    incrArray = 0;
                    if (result == TCL_ERROR) goto end_of_else;
                    UpdateBlockMinMaxStackDepth();
                    break;
                    
                case ASSEM_STR_EQ:
                    TclEmitOpcode(INST_STR_EQ, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_STR_NEQ:
                    TclEmitOpcode(INST_STR_NEQ, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_STR_CMP:
                    TclEmitOpcode(INST_STR_CMP, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_STR_LEN:
                    TclEmitOpcode(INST_STR_LEN, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_STR_INDEX:
                    TclEmitOpcode(INST_STR_INDEX, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_STR_MATCH:
                    opnd = atoi(instArgs[0]);
                    if (opnd > 1) opnd = 1;
                    if (opnd < 0) opnd = 0;
            		TclEmitInstInt1(INST_STR_MATCH, opnd,	envPtr);
            		UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_LIST_INDEX:
                    TclEmitOpcode(INST_LIST_INDEX, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                case ASSEM_LIST_LENGTH:
                    TclEmitOpcode(INST_LIST_LENGTH, envPtr);
                    UpdateBlockMinMaxStackDepth();
                    break;
                
                case ASSEM_OVER:
                    opnd = atoi(instArgs[0]);
            	    TclEmitInstInt4(INST_OVER, opnd, envPtr);
            	    UpdateBlockMinMaxStackDepth();
                    break;
            	    
                case ASSEM_LSETLIST:
                	TclEmitOpcode(INST_LSET_LIST, envPtr);
                	UpdateBlockMinMaxStackDepth();
                    break;
                                
                default:
                    printf("Instruction \"%s\" could not be found, end of switch, default label\n", instName);
                    return TCL_ERROR;
            }
            
         }
         end_of_else:
         result = CheckStack(head_bb);
         Tcl_DeleteHashTable(LabelHashPtr); // Actually, we need to free each label as well.
         return result;
     }
}

int TclCheckByteCodeSyntax(
    Tcl_Interp *interp, 
    Tcl_Token *tokenPtr) {
    int index,indexinner, length, result, strLength;
    Tcl_Obj *bcList, *bc, *bcArg, *Op, *elementc, *elementv;
    char * bcStr;

    bcList = Tcl_NewStringObj(tokenPtr[1].start, tokenPtr[1].size);

    if (Tcl_ListObjLength(interp, bcList, &length) == TCL_ERROR) {
        return TCL_ERROR;
    }
    
    if (length == 0) {
        // Don't know what to do yet
        return TCL_OK;
    }
    
    for (index = 0; index < length; index++) {
        result = Tcl_ListObjIndex(interp, bcList, index, &bc);
        if (result == TCL_ERROR) 
            return TCL_ERROR;
            
        for (indexinner =0; indexinner < 2; indexinner++) {
            result = Tcl_ListObjIndex(interp, bc, indexinner, &Op);
            if (result == TCL_ERROR) 
                return TCL_ERROR;        
        }
    }
    return TCL_OK;
}


static basicBlock * AllocBB(unsigned char * start,
    int block_depth,
    int initialStackDepth) 
{
    basicBlock * bb = (basicBlock *) ckalloc(sizeof(basicBlock));
    bb->start = start;
    bb->visited = 0;
    bb->block_depth = block_depth;
    bb->initialStackDepth = initialStackDepth;
    bb->predecessor = NULL;
    bb->jumpTargetLabelHashEntry = NULL;
    bb->successor1 = NULL;
    return bb;
}

static int
CheckStack(basicBlock * head) {
    StackCheckerState * st = (StackCheckerState *) ckalloc(sizeof(StackCheckerState));
    st->maxDepth = 0;
    if(StackCheckBasicBlock(st, head, NULL, 0) == TCL_ERROR) {
        printf("Stack balance check failed\n");
        return TCL_ERROR;
    }
}

static int
StackCheckBasicBlock(StackCheckerState *st, basicBlock * blockPtr, basicBlock * predecessor, int initialStackDepth) { 
    if (blockPtr->visited) {
        if (blockPtr->initialStackDepth != initialStackDepth) {
            printf("Inconsistent initialStackDepth\n");
            /* Trace the offending basicBlock */
            return TCL_ERROR;
        }
        else {
            return TCL_OK;
        }
    }
    else {
        blockPtr->visited = 1;
        blockPtr->predecessor = predecessor;
        blockPtr->initialStackDepth = initialStackDepth;
        if (initialStackDepth + blockPtr->minStackDepth < 0) {
            printf("Minstackdepth underflows the stack\n");
            return TCL_ERROR;
        }
        if (initialStackDepth + blockPtr->maxStackDepth > st->maxDepth) {
            st->maxDepth = initialStackDepth + blockPtr->maxStackDepth;
        }
    }
    int stackDepth = initialStackDepth + blockPtr->finalStackDepth;
    int result = TCL_OK;
    if (blockPtr->successor1 != NULL) {
        result = StackCheckBasicBlock(st, blockPtr->successor1, blockPtr, stackDepth);        
        
    }
    if (result == TCL_OK && blockPtr->jumpTargetLabelHashEntry != NULL) {
        basicBlock * targetBlock = (basicBlock *) Tcl_GetHashValue(blockPtr->jumpTargetLabelHashEntry);
        result = StackCheckBasicBlock(st, targetBlock, blockPtr, stackDepth);
    }
    return result;
    
}

