#ifndef _TCL_ASSEMBLY
#define _TCL_ASSEMBLY 1

#include "tclCompile.h"

typedef struct StackCheckerState { 
    int maxDepth;
} StackCheckerState;

typedef struct basicBlock {
    unsigned char * start;
    unsigned char * jumpTarget;
    unsigned char * nextBlock;
    int may_fall_thru;      /* ends on conditional jump or has no jumps */
    int visited;             
    int block_depth;
    int script_depth;
    int startBytecodeIndex;            /* Absolute bytecode index of the start of the block */
    struct BasicBlock* predecessor; /* Temporary that identifies the
                                         * predecessor of this block in the
                                         * spanning tree */
    struct basicBlock * successor1; /* BasicBlock structure of the following block : 
                                NULL at the end of the bytecode sequence or 
                                if the block ends in an unconditional jump */ 
    Tcl_HashEntry * jumpTargetLabelHashEntry;  /* Jump target label if the jump target is unresolved */
    
    int initialStackDepth;
    int minStackDepth;
    int maxStackDepth; 
    int finalStackDepth;

} basicBlock;

typedef enum talInstType {
    ASSEM_PUSH,     /* These instructions will be looked up from talInstructionTable */
    ASSEM_1BYTE,    /* The instructions that are directly mapped to tclInstructionTable in tclCompile.c*/
    ASSEM_LOAD,     /* Load instruction  */
    ASSEM_LOAD_STK,
    ASSEM_LOAD_ARR,  /* Load a value from an array index  */
    ASSEM_LOAD_ARR_STK,
    ASSEM_STORE,         /* Store instructions */
    ASSEM_STORE_STK,
    ASSEM_STORE_ARR,
    ASSEM_STORE_ARR_STK,
    ASSEM_INCR,
    ASSEM_INCR_IMM,
    ASSEM_INCR_STK,
    ASSEM_INCR_STK_IMM,
    ASSEM_INCR_ARRAY,
    ASSEM_INCR_ARRAY_IMM,
    ASSEM_INCR_ARRAY_STK,
    ASSEM_INCR_ARRAY_STK_IMM,
    ASSEM_STR_EQ,
    ASSEM_STR_NEQ,
    ASSEM_STR_CMP,
    ASSEM_STR_LEN,
    ASSEM_STR_INDEX,
    ASSEM_STR_MATCH,
    ASSEM_LIST_INDEX,
    ASSEM_LIST_LENGTH,
    ASSEM_OVER,
    ASSEM_LSETLIST,
    ASSEM_LABEL,
    ASSEM_JUMP,
    ASSEM_JUMP_FALSE,
    ASSEM_JUMP_TRUE    
} talInstType;

typedef struct talInstDesc {
    const char *name;		/* Name of instruction. */
    int stackEffect;
    talInstType instType; 	/* The type of instruction */
    int tclInstCode;
    int operandsConsumed;
    int operandsProduced;

} talInstDesc;

typedef struct label {
    int isDefined;
    unsigned char * offset;    
} label;


#endif 